/* mb.c     91/03/21   builtin predicates                                     */

#include "mini.h"

#ifdef Trace

long    trace_instr[MaxTrace+BuiltinSize];
int     argfreeze;
int     argcount;
int     repush;

#endif

MWPTR *ctbottom;

static int compute(MWPTR result, MW arg);
static int term_equal(MWPTR arg1ptr, MWPTR arg2ptr);
static int term_less(MWPTR arg1ptr, MWPTR arg2ptr);

#ifdef Atari
extern long Super(void *stack);
#else
extern long clock(void);
#endif

int b_cputime(MW arg[])
{
MW      m;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (NotRefUndef(m))
	return(ErrorNumber = BuiltinInstError);
Trail(mptr);

#ifdef Atari

	{
	long    *timer, oldstack;
	
	timer = (long *) 0x4BA;
	oldstack = Super(0L);
	*mptr = MakeInt((*timer) * 5);
	(void) Super((void *) oldstack);
	}

#else

*mptr = MakeInt(clock() / 1000);

#endif

return(Success);
}


int b_fail()
{
return(Fail);
}


int b_atom(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Atom(m))
	return(Success);
return(Fail);
}


int b_atomic(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if ((Int(m)) || (Atom(m)))
	return(Success);
return(Fail);
}


int b_compound(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (ListOrStruct(m))
	return(Success);
return(Fail);
}


int b_integer(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Int(m))
	return(Success);
return(Fail);
}


int b_list(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (List(m))
	return(Success);
return(Fail);
}


int b_nil(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Nil(m))
	return(Success);
return(Fail);
}


int b_nonvar(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (NotRefUndef(m))
	return(Success);
return(Fail);
}


int b_number(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Int(m))
	return(Success);
return(Fail);
}


int b_struct(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Struct(m))
	return(Success);
return(Fail);
}


int b_var(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (RefUndef(m))
	return(Success);
return(Fail);
}


int b_int_eq(MW arg[])
{
MW      m1, m2;

if (compute(&m1, arg[0]) != Success)
	return(ErrorNumber);
if (compute(&m2, arg[1]) != Success)
	return(ErrorNumber);
if (XtrInt(m1) == XtrInt(m2))
	return(Success);
return(Fail);
}


int b_int_ne(MW arg[])
{
MW      m1, m2;

if (compute(&m1, arg[0]) != Success)
	return(ErrorNumber);
if (compute(&m2, arg[1]) != Success)
	return(ErrorNumber);
if (XtrInt(m1) != XtrInt(m2))
	return(Success);
return(Fail);
}


int b_int_lt(MW arg[])
{
MW      m1, m2;

if (compute(&m1, arg[0]) != Success)
	return(ErrorNumber);
if (compute(&m2, arg[1]) != Success)
	return(ErrorNumber);
if (XtrInt(m1) < XtrInt(m2))
	return(Success);
return(Fail);
}


int b_int_le(MW arg[])
{
MW      m1, m2;

if (compute(&m1, arg[0]) != Success)
	return(ErrorNumber);
if (compute(&m2, arg[1]) != Success)
	return(ErrorNumber);
if (XtrInt(m1) <= XtrInt(m2))
	return(Success);
return(Fail);
}


int b_int_gt(MW arg[])
{
MW      m1, m2;

if (compute(&m1, arg[0]) != Success)
	return(ErrorNumber);
if (compute(&m2, arg[1]) != Success)
	return(ErrorNumber);
if (XtrInt(m1) > XtrInt(m2))
	return(Success);
return(Fail);
}


int b_int_ge(MW arg[])
{
MW      m1, m2;

if (compute(&m1, arg[0]) != Success)
	return(ErrorNumber);
if (compute(&m2, arg[1]) != Success)
	return(ErrorNumber);
if (XtrInt(m1) >= XtrInt(m2))
	return(Success);
return(Fail);
}


int b_is(MW arg[])
{
MW      m, result;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	return(compute(mptr, arg[1]));
	}
if (compute(&result, arg[1]) != Success)
	return(ErrorNumber);
if (result == m)
	return(Success);
return(Fail);
}


static int compute(MWPTR resultptr, MW arg)
{
MW      m1, m2, arity, op;
MWPTR   argptr;

ValDeref(arg);
if (Int(arg))
	{
	*resultptr = arg;
	return(Success);
	}
if (NotStruct(arg))
	if (RefUndef(arg))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
argptr = XtrStruct(arg);
op = FunctIndex(XtrFunctor(argptr));
arity = XtrArity(argptr);
argptr = FirstArgPtr(argptr);
if (compute(&m1, *argptr) != Success)
	return(ErrorNumber);
if (arity == 1)
	{
	m2 = XtrInt(m1);
	if (op == PlusAtomIndex)
		{
		*resultptr = m1;
		return(Success);
		}
	else if (op == MinusAtomIndex)
		{
		*resultptr = MakeInt(-(m2));
		return(Success);
		}
	else if (op == AbsAtomIndex)
		{
		if (m2 < 0)
			*resultptr = MakeInt(-m2);
		else
			*resultptr = m1;
		return(Success);
		}
	else if (op == BitNotAtomIndex)
		{
		*resultptr = MakeInt(~(m2));
		return(Success);
		}
	else
		return(ErrorNumber = BuiltinTypeError);
	}
else if (arity != 2)
	return(ErrorNumber = BuiltinTypeError);
if (compute(&m2, *(argptr+1)) != Success)
	return(ErrorNumber);
m1 = XtrInt(m1);
m2 = XtrInt(m2);
if (op == PlusAtomIndex)
	{
	*resultptr = MakeInt(m1 + m2);
	return(Success);
	}
else if (op == MinusAtomIndex)
	{
	*resultptr = MakeInt(m1 - m2);
	return(Success);
	}
else if (op == MulAtomIndex)
	{
	*resultptr = MakeInt(m1 * m2);
	return(Success);
	}
else if (op == DivAtomIndex)
	{
	*resultptr = MakeInt(m1 / m2);
	return(Success);
	}
else if (op == ModAtomIndex)
	{
	*resultptr = MakeInt(m1 % m2);
	return(Success);
	}
else if (op == ShiftRightAtomIndex)
	{
	*resultptr = MakeInt(m1 >> m2);
	return(Success);
	}
else if (op == ShiftLeftAtomIndex)
	{
	*resultptr = MakeInt(m1 << m2);
	return(Success);
	}
else if (op == BitAndAtomIndex)
	{
	*resultptr = MakeInt(m1 & m2);
	return(Success);
	}
else if (op == BitOrAtomIndex)
	{
	*resultptr = MakeInt(m1 | m2);
	return(Success);
	}
return(ErrorNumber = BuiltinTypeError);
}


int b_dec(MW arg[])
{
MW      m;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (NotRefUndef(m))
	return(ErrorNumber = BuiltinInstError);
m = arg[1];
ValDeref(m);
if (NotInt(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
Trail(mptr);
*mptr = MakeInt(XtrInt(m)-1);
return(Success);
}


int b_inc(MW arg[])
{
MW      m;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (NotRefUndef(m))
	return(ErrorNumber = BuiltinInstError);
m = arg[1];
ValDeref(m);
if (NotInt(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
Trail(mptr);
*mptr = MakeInt(XtrInt(m)+1);
return(Success);
}


int b_arg(MW arg[])
{
MW      m, i;
MWPTR   mptr;

m = arg[0];
ValDeref(m);
if (NotInt(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
i = XtrInt(m);
m = arg[1];
ValDeref(m);
if (NotStruct(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
mptr = XtrStruct(m);
if ((i <= 0) || (i > XtrArity(mptr)))
	return(ErrorNumber = BuiltinRangeError);
return(unify(arg+2, FirstArgPtr(mptr) + i - 1));
}


int b_functor(MW arg[])
{
MW      m, mi, i;
MWPTR   mptr, functor;

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	m = arg[2];
	ValDeref(m);
	if (NotInt(m))
		if (RefUndef(m))
			return(ErrorNumber = BuiltinInstError);
		else
			return(ErrorNumber = BuiltinTypeError);
	i = XtrInt(m);
	if (i < 0)
		return(ErrorNumber = BuiltinRangeError);
	m = arg[1];
	ValDeref(m);
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	if (i == 0)
		{
		if (ListOrStruct(m))
			return(ErrorNumber = BuiltinTypeError);
		*mptr = m;
		return(Success);
		}
	if (NotAtom(m))
		return(ErrorNumber = BuiltinTypeError);
	if ((i == 2) && (m == DotAtom))
		{
		CreateList(mptr);
		}
	else
		{
		if ((functor = find_func(XtrAtomIndex(m), i)) == 0)
			return(ErrorNumber);
		CreateStruct(mptr, functor, i);
		}
	while (--i >= 0)
		*mptr++ = MakeUndef(mptr);
	return(Success);
	}
if (Struct(m))
	{
	mptr = XtrStruct(m);
	functor = XtrFunctor(mptr);
	m = MakeAtom(FunctIndex(functor));
	i = FunctArity(functor);
	}
else if (List(m))
	{
	m = DotAtom;
	i = 2;
	}
else
	i = 0;
mi = *(mptr = arg + 2);
StructDeref(mi, mptr);
if (RefUndef(mi))
	{
	Trail(mptr);
	*mptr = MakeInt(i);
	}
else if (Int(mi))
	{
	mi = XtrInt(mi);
	if (mi < 0)
		return(ErrorNumber = BuiltinRangeError);
	if (i != mi)
		return(Fail);
	}
else
	return(ErrorNumber = BuiltinInstError);
return(unify(arg + 1, &m));
}


int b_univ(MW arg[])
{
MW      m1, m2, i;
MWPTR   m1ptr, m2ptr, m3ptr;
int     err;

m1 = *(m1ptr = arg);
StructDeref(m1, m1ptr);
m2 = *(m2ptr = arg + 1);
StructDeref(m2, m2ptr);
if (Val(m1))
	{
	if (ListOrStruct(m1))
		if (List(m1))
			{
			m1ptr = XtrList(m1);
			TraceInstr(t_xtrlist)
			m1 = DotAtom;
			i = 2;
			}
		else
			{
			m1ptr = XtrStruct(m1);
			m3ptr = XtrFunctor(m1ptr);
			m1 = MakeAtom(FunctIndex(m3ptr));
			i = FunctArity(m3ptr);
			m1ptr = FirstArgPtr(m1ptr);
			}
	else
		i = 0;
	if (RefUndef(m2))
		{
		Trail(m2ptr);
		CreateList(m2ptr);
		*m2ptr++ = m1;
		}
	else if (List(m2))
		{
		m2ptr = (m3ptr = XtrList(m2)) + 1;
		TraceInstr(t_xtrlist)
		m2 = *m3ptr;
		StructDeref(m2, m3ptr);
		if (RefUndef(m2))
			{
			Trail(m3ptr);
			*m3ptr = m1;
			}
		else if (m1 != m2)
			return(Fail);
		m2 = *m2ptr;
		StructDeref(m2, m2ptr);
		}
	else
		return(ErrorNumber = BuiltinTypeError);
	while (List(m2) && (--i >= 0))
		{
		m2ptr = XtrList(m2);
		TraceInstr(t_xtrlist)
		if ((err = unify(m1ptr, m2ptr)) != Success)
			return(err);
		m1ptr++;
		m2 = *(++m2ptr);
		StructDeref(m2, m2ptr);
		}
	if (i <= 0)
		{
		if (RefUndef(m2))
			{
			*m2ptr = MakeNil;
			return(Success);
			}
		if (Nil(m2))
			return(Success);
		if (List(m2))
			return(Fail);
		return(ErrorNumber = BuiltinTypeError);
		}
	if (NotRefUndef(m2))
		return(Fail);
	Trail(m2ptr);
	while (--i >= 0)
		{
		CreateList(m2ptr);
#ifdef UndefSelfRef
		*m2ptr++ = *m1ptr++;
#else
		if (Undef(*m2ptr++ = *m1ptr++))
			*(m2ptr-1) = MakeRef(m1ptr - 1);
#endif /* UndefSelfRef */
		}
	*m2ptr = MakeNil;
	return(Success);
	}
if (RefUndef(m2))
	return(ErrorNumber = BuiltinInstError);
if (NotList(m2))
	return(ErrorNumber = BuiltinTypeError);
m1 = *(m2ptr = XtrList(m2));
TraceInstr(t_xtrlist)
ValDeref(m1);
if (RefUndef(m1))
	return(ErrorNumber = BuiltinInstError);
if (ListOrStruct(m1))
	return(ErrorNumber = BuiltinTypeError);
i = 0;
m2 = *(++m2ptr);
ValDeref(m2);
while (List(m2))
	{
	i++;
	m2 = *(XtrList(m2) + 1);
	TraceInstr(t_xtrlist)
	ValDeref(m2);
	}
if (RefUndef(m2))
	return(ErrorNumber = BuiltinInstError);
if (NotNil(m2))
	return(ErrorNumber = BuiltinTypeError);
Trail(m1ptr);
if (i == 0)
	{
	*m1ptr = m1;
	return(Success);
	}
/* assert (i > 0) */
if (NotAtom(m1))
	return(ErrorNumber = BuiltinTypeError);
if ((i == 2) && (m1 == DotAtom))
	{
	CreateList(m1ptr);
	}
else
	{
	if ((m3ptr = find_func(XtrAtomIndex(m1), i)) == 0)
		return(ErrorNumber);
	CreateStruct(m1ptr, m3ptr, i);
	}
while (--i >= 0)
	{
	m2 = *m2ptr;
	ValDeref(m2);
	m2ptr = XtrList(m2);
	TraceInstr(t_xtrlist)
#ifdef UndefSelfRef
	*m1ptr++ = *m2ptr++;
#else
	if (Undef(*m1ptr++ = *m2ptr++))
		*(m1ptr-1) = MakeRef(m2ptr-1);
#endif /* UndefSelfRef */
	}
return(Success);
}


int var_copy_term(MWPTR m1ptr, MWPTR m2ptr)
{
MWPTR *ctptr, functor;
MW    m1, arity;

m1 = *m1ptr;
StructDeref(m1, m1ptr);
if (RefUndef(m1))
	{
	ctptr = trail;
	while ((ctptr -= 2) >= ctbottom) 
		if (*ctptr == m1ptr)
			{
			*m2ptr = MakeRef(*(ctptr + 1));
			return Success;
			}
	*m2ptr = MakeUndef(m2ptr);
	if ((MWPTR) (trail + 2) >= copyptr)
		return(ErrorNumber = CopyOverflowError);
	*trail++ = m1ptr;
	*trail++ = m2ptr;
	return(Success);
	}
if (List(m1))
	{
	m1ptr = XtrList(m1);
	CreateList(m2ptr);
	if (var_copy_term(m1ptr, m2ptr) != Success)
		return(ErrorNumber);
	return(var_copy_term(m1ptr+1, m2ptr+1));
	}
if (Struct(m1))
	{
	m1ptr = XtrStruct(m1);
	functor = XtrFunctor(m1ptr);
	arity = FunctArity(functor);
	CreateStruct(m2ptr, functor, arity);
	m1ptr = FirstArgPtr(m1ptr);
	while (--arity >= 0)
		if (var_copy_term(m1ptr++, m2ptr++) != Success)
			return(ErrorNumber);
	}
else
	*m2ptr = m1;
return(Success);
}


int copy_term(MWPTR m1ptr, MWPTR m2ptr)
{
int result;

ctbottom = trail;
result = var_copy_term(m1ptr, m2ptr);
trail = ctbottom;
return(result);
}


int b_copy_term(MW arg[])
{
MW m1;
MWPTR m0ptr, m1ptr;

m1 = *(m1ptr = arg + 1);
StructDeref(m1, m1ptr);
if (RefUndef(m1))
	{
	m1 = arg[0];
	ValDeref(m1);
	if (RefUndef(m1))
		return(Success);
	Trail(m1ptr);
	m0ptr = m1ptr;
	}
else
	{
	MakeCopyUndef;
	m0ptr = copyptr;
	}
if (copy_term(arg, m0ptr) != Success)
	return(ErrorNumber);
return(unify(m0ptr, m1ptr));
}


int b_atom_length(MW arg[])
{
MW      m, l;
MWPTR   mptr;
CHARPTR cptr;

m = arg[0];
ValDeref(m);
if (NotAtom(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
cptr = XtrAtomPtr(m);
l = 0;
while (*cptr++)
	l++;
l = MakeInt(l);
m = *(mptr = arg + 1);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	*mptr = l;
	return(Success);
	}
if (Int(m))
	if (l == m)
		return(Success);
	else
		return(Fail);
return(ErrorNumber = BuiltinTypeError);
}


int b_atom_list(MW arg[])
{
MW      m, l;
MWPTR   mptr, lptr;
CHARPTR aptr, cptr;
char    string[2];

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	l = arg[1];
	ValDeref(l);
	Trail(mptr);
	aptr = (CHARPTR) trail;
	while(List(l))
		{
		lptr = XtrList(l);
		l = *lptr;
		ValDeref(l);
		if (NotAtom(l))
			return(ErrorNumber = BuiltinTypeError);
		cptr = XtrAtomPtr(l);
		if (((*aptr++ = *cptr) == 0) || (*(cptr+1) != 0))
			return(ErrorNumber = BuiltinTypeError);
		l = *(lptr+1);
		ValDeref(l);
		}
	if (RefUndef(l))
		return(ErrorNumber = BuiltinInstError);
	if (NotNil(l))
		return(ErrorNumber = BuiltinTypeError);
	*aptr = 0;
	return(find_atom((CHARPTR) trail, mptr));
	}
if (Atom(m))
	{
	aptr = XtrAtomPtr(m);
	lptr = arg + 1;
	string[1] = 0;
	while (*aptr)
		{
		l = *lptr;
		StructDeref(l, lptr);
		if (RefUndef(l))
			{
			Trail(lptr);
			CreateList(lptr);
			mptr = lptr++;
			m = MakeUndef(mptr);
			*lptr = MakeUndef(lptr);
			}
		else if (List(l))
			{
			m = *(mptr = XtrList(l));
			lptr = mptr + 1;
			StructDeref(m, mptr);
			}
		else if (Nil(l))
			return(Fail);
		else
			return(ErrorNumber = BuiltinTypeError);
		if RefUndef(m)
			{
			string[0] = *aptr++;
			Trail(mptr);
			if (find_atom(string, mptr) != Success)
				return(ErrorNumber);
			}
		else if (Atom(m))
			{
			cptr = XtrAtomPtr(m);
			if ((*cptr == 0) || (*(cptr+1) != 0))
				return(ErrorNumber = BuiltinTypeError);
			if (*aptr++ != *cptr)
				return(Fail);
			}
		else
			return(ErrorNumber = BuiltinTypeError);
		}
	l = *lptr;
	StructDeref(l, lptr);
	if (RefUndef(l))
		{
		Trail(lptr);
		*lptr = MakeNil;
		return(Success);
		}
	if (Nil(l))
		return(Success);
	}
return(ErrorNumber = BuiltinTypeError);
}


int b_char_int(MW arg[])
{
MW      m, l;
MWPTR   mptr;
CHARPTR cptr;
char    string[2];

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	l = arg[1];
	ValDeref(l);
	if (NotInt(l))
		if (RefUndef(l))
			return(ErrorNumber = BuiltinInstError);
		else
			return(ErrorNumber = BuiltinTypeError);
	l = XtrInt(l);
	if ((l < 0) || (l > 255))
		return(ErrorNumber = BuiltinTypeError);
	string[0] = l;
	string[1] = 0;
	Trail(mptr);
	return(find_atom(string, mptr));
	}
if (Atom(m))
	{
	cptr = XtrAtomPtr(m);
	l = *cptr++;
	if ((l == 0) || (*cptr != 0))
		return(ErrorNumber = BuiltinTypeError);
	l = MakeInt(l);
	m = *(mptr = arg + 1);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		Trail(mptr);
		*mptr = l;
		return(Success);
		}
	if (Int(m))
		if (l == m)
			return(Success);
		else
			return(Fail);
	}
return(ErrorNumber = BuiltinTypeError);
}


int struct_unify(MW m1, MW m2)
{
MW      i;
MWPTR   m1ptr, m2ptr;
MWPTR   m3ptr, m4ptr;

if (List(m1))
	if (List(m2))
		{
		m3ptr = XtrList(m1);
		m4ptr = XtrList(m2);
		TraceInstr(t_xtrlist)
		TraceInstr(t_xtrlist)
		i = 2;
		}
	else
		return(FALSE);
else
	if (Struct(m2))
		{
		m3ptr = XtrStruct(m1);
		m4ptr = XtrStruct(m2);
		i = XtrArity(m3ptr);
		StructHeadUnify(m3ptr, m4ptr);
		}
	else
		return(FALSE);

while (--i >= 0)
	{
	if (NotVal(m1 = *(m1ptr = m3ptr++)))
		{
		StructDeref(m1, m1ptr);
		if (RefUndef(m1))
			{
			if (NotVal(m2 = *(m2ptr = m4ptr++)))
				{
				StructDeref(m2, m2ptr);
				if (RefUndef(m2))
					if (Pointsto(m1ptr, m2ptr))
						{
						TraceInstr(t_makeref)
						AssignCont(m1ptr, MakeRef(m2ptr))
						}
					else if (Pointsto(m2ptr, m1ptr))
						{
						TraceInstr(t_makeref)
						AssignCont(m2ptr, MakeRef(m1ptr))
						}
					else
						continue;
				}
			AssignCont(m1ptr, m2)
			}
		}
	if (NotVal(m2 = *(m2ptr = m4ptr++)))
		{
		StructDeref(m2, m2ptr);
		if (RefUndef(m2))
			AssignCont(m2ptr, m1)
		}
	if (m1 == m2)
		continue;
	if (ListOrStruct(m1) &&	struct_unify(m1, m2))
		continue;
	return(FALSE);
	}
return(TRUE);
}


int unify(MWPTR m1ptr, MWPTR m2ptr)
{
MW m1, m2;

if (NotVal(m1 = *m1ptr))
	{
	StructDeref(m1, m1ptr);
	if (RefUndef(m1))
		{
		if (NotVal(m2 = *m2ptr))
			{
			StructDeref(m2, m2ptr);
			if (RefUndef(m2))
				if (Pointsto(m1ptr, m2ptr))
					{
					TraceInstr(t_makeref)
					AssignSuccess(m1ptr, MakeRef(m2ptr))
					}
				else if (Pointsto(m2ptr, m1ptr))
					{
					TraceInstr(t_makeref)
					AssignSuccess(m2ptr, MakeRef(m1ptr))
					}
				else
					return(Success);
			}
		AssignSuccess(m1ptr, m2)
		}
	}
if (NotVal(m2 = *m2ptr))
	{
	StructDeref(m2, m2ptr);
	if (RefUndef(m2))
		AssignSuccess(m2ptr, m1)
	}
if (m1 == m2)
	return(Success);
if (ListOrStruct(m1) &&	struct_unify(m1, m2))
	return(Success);
return(Fail);
}


int b_exit_true()
{
return(ErrorNumber = ExitTrueError);
}


int b_exit_fail()
{
return(ErrorNumber = ExitFailError);
}


int b_block()
{
return(Success);
}


int b_exit_block()
{
return(Success);
}


int b_unify(MW arg[])
{
return(unify(arg, arg + 1));
}


int b_notunify(MW arg[])
{
int     result;
MWPTR   *trailsave, tptr /* , lastcopysave */;

trailsave = trail;
/* lastcopysave = lastcopy;
lastcopy = copyptr; */
result = unify(arg, arg + 1);
/* lastcopy = lastcopysave; */
while (trail > trailsave)
	{
	tptr = *(--trail);
	*tptr = MakeUndef(tptr);
	}
if (result == Success)
	return(Fail);
if (result == Fail)
	return(Success);
return(result);
}


static int struct_equal(MW m1, MW m2)
{
MW      i;
MWPTR   m1ptr, m2ptr;
MWPTR   m3ptr, m4ptr;

if (List(m1))
	if (List(m2))
		{
		m3ptr = XtrList(m1);
		m4ptr = XtrList(m2);
		TraceInstr(t_xtrlist)
		TraceInstr(t_xtrlist)
		i = 2;
		}
	else
		return(FALSE);
else
	if (Struct(m2))
		{
		m3ptr = XtrStruct(m1);
		m4ptr = XtrStruct(m2);
		i = XtrArity(m3ptr);
		StructHeadUnify(m3ptr, m4ptr);
		}
	else
		return(FALSE);

while (--i >= 0)
	{
	if (NotVal(m1 = *(m1ptr = m3ptr++)))
		{
		StructDeref(m1, m1ptr);
		if (RefUndef(m1))
			{
			m2 = *(m2ptr = m4ptr++);
			StructDeref(m2, m2ptr);
			if (RefUndef(m2) && (m1ptr == m2ptr))
				continue;
			return(FALSE);
			}
		}
	if (NotVal(m2 = *m4ptr++))
		{
		ValDeref(m2);
		if (RefUndef(m2))
			return(FALSE);
		}
	if (m1 == m2)
		continue;
	if (ListOrStruct(m1) && struct_equal(m1, m2))
		continue;
	return(FALSE);
	}
return(TRUE);
}


static int term_equal(MWPTR m1ptr, MWPTR m2ptr)
{
MW      m1, m2;

if (NotVal(m1 = *m1ptr))
	{
	StructDeref(m1, m1ptr);
	if (RefUndef(m1))
		{
		m2 = *m2ptr;
		StructDeref(m2, m2ptr);
		if (RefUndef(m2) && (m1ptr == m2ptr))
			return(Success);
		return(Fail);
		}
	}
if (NotVal(m2 = *m2ptr))
	{
	ValDeref(m2);
	if (RefUndef(m2))
		return(Fail);
	}
if (m1 == m2)
	return(Success);
if (ListOrStruct(m1) && struct_equal(m1, m2))
	return(Success);
return(Fail);
}


int b_equal(MW arg[])
{
return(term_equal(arg, arg + 1));
}


int b_notequal(MW arg[])
{
if (term_equal(arg, arg + 1) == Success)
	return(Fail);
return(Success);
}


static int term_less(MWPTR arg1ptr, MWPTR arg2ptr)
{
MW  arg1, arg2;
MW  arity, arity2;
int result;

arg1 = *arg1ptr;
StructDeref(arg1, arg1ptr);
arg2 = *arg2ptr;
StructDeref(arg2, arg2ptr);
if (RefUndef(arg1))
	if (RefUndef(arg2))
		if (Pointsto(arg2ptr, arg1ptr))
			return(Success);
		else
			return(Fail);
	else
		return(Success);
if (Atom(arg1))
	if (Atom(arg2))
		if (strcmp(XtrAtomPtr(arg1), XtrAtomPtr(arg2)) < 0)
			return(Success);
		else
			return(Fail);
	else if (Nil(arg2))
		if (strcmp(XtrAtomPtr(arg1), "[]") < 0)
			return(Success);
		else
			return(Fail);
	else if (ListOrStruct(arg2))
		return(Success);
	else
		return(Fail);
if (Int(arg1))
	if (Int(arg2))
		if (XtrInt(arg1) < XtrInt(arg2))
			return(Success);
		else
			return(Fail);
	else if (RefUndef(arg2))
		return(Fail);
	else
		return(Success);
if (Nil(arg1))
	if (Atom(arg2))
		if (strcmp("[]", XtrAtomPtr(arg2)) < 0)
			return(Success);
		else
			return(Fail);
	else if (ListOrStruct(arg2))
		return(Success);
	else
		return(Fail);
if (List(arg1))
	if (List(arg2))
		{
		arity = 2;
		arg1ptr = XtrList(arg1);
		arg2ptr = XtrList(arg2);
		}
	else if (Struct(arg2))
		{
		arg2ptr = XtrStruct(arg2);
		arity = XtrArity(arg2ptr);
		if (arity > 2)
			return(Success);
		if (arity < 2)
			return(Fail);
		result = strcmp(".", IndexPtr(FunctIndex(XtrFunctor(arg2ptr))));
		if (result  < 0)
			return(Success);
		if (result  > 0)
			return(Fail);
		arg1ptr = XtrList(arg1);
		arg2ptr = FirstArgPtr(arg2ptr);
		}
	else
		return(Fail);
else
	if (List(arg2))
		{
		arg1ptr = XtrStruct(arg1);
		arity = XtrArity(arg1ptr);
		if (arity < 2)
			return(Success);
		if (arity > 2)
			return(Fail);
		result = strcmp(".", IndexPtr(FunctIndex(XtrFunctor(arg1ptr))));
		if (result  > 0)
			return(Success);
		if (result  < 0)
			return(Fail);
		arg1ptr = FirstArgPtr(arg1ptr);
		arg2ptr = XtrList(arg2);
		}
	else if (Struct(arg2))
		{
		arg1ptr = XtrStruct(arg1);
		arity = XtrArity(arg1ptr);
		arg2ptr = XtrStruct(arg2);
		arity2 = XtrArity(arg2ptr);
		if (arity < arity2)
			return(Success);
		if (arity > arity2)
			return(Fail);
		result = strcmp(IndexPtr(FunctIndex(XtrFunctor(arg1ptr))),
		                IndexPtr(FunctIndex(XtrFunctor(arg2ptr))));
		if (result  < 0)
			return(Success);
		if (result  > 0)
			return(Fail);
		arg1ptr = FirstArgPtr(arg1ptr);
		arg2ptr = FirstArgPtr(arg2ptr);
		}
	else
		return(Fail);
while ((--arity >= 0) && (term_equal(arg1ptr, arg2ptr) == Success))
	{
	arg1ptr++;
	arg2ptr++;
	}
if (arity < 0)
	return(Fail);
return(term_less(arg1ptr, arg2ptr));
}


int b_less(MW arg[])
{
return(term_less(arg, arg + 1));
}


int b_greater(MW arg[])
{
return(term_less(arg + 1, arg));
}


int b_greaterequal(MW arg[])
{
if (term_less(arg, arg + 1) == Success)
	return(Fail);
return(Success);
}


int b_equalless(MW arg[])
{
if (term_less(arg + 1, arg) == Success)
	return(Fail);
return(Success);
}
