/* mc.c     91/03/26   intermediate code compiler and decompiler              */

#include "mini.h"

#define MAXVARS     256
#define MAXTMPS     16

#define VOID_VAR    0
#define TEMP_VAR    1
#define LOCAL_VAR   2

#define Var(cp)     loc_vars[-3-(SopFetch(cp) / (SIZEMW / SIZECHAR))]
#define ShortVar(o) loc_vars[o]
#define Tmp(cp)     tmp_vars[SopFetch(cp) / (SIZEMW / SIZECHAR)]
#define ShortTmp(o) tmp_vars[o]
#define Offset(o)   ((-3-(o))*(SIZEMW / SIZECHAR))

#define CodeStackCheck  if(codeptr>=(CODEPTR)StackPtr) \
                            longjmp(compiler_err, OutOfMemoryError)


typedef struct {
	MWPTR       address;        /* address of free variable                 */
	SHORT       type;           /* void or temporary or local               */
	SHORT       offset;         /* offset for local variables               */
	SHORT       var_count;      /* occurence count of variable in a clause  */
	SHORT       var_nr;         /* current occurene number of variable      */
	SHORT       first_goal;     /* goal number of first occurence, head = 0 */
	SHORT       last_goal;      /* goal number of last occurence            */
} VAR_STRUCT;

typedef struct {
	SHORT       goal_count;     /* count of goals in a clause               */
	SHORT       headgoal_count; /* count of head temp goals in a clause     */
	SHORT       goal_nr;        /* currenct number of goal                  */
	SHORT       all_vars;       /* count of all variables                   */
	SHORT       head_vars;      /* count of head variables                  */
	SHORT       frame_size;     /* (count of local variables + 2) * SIZEMW  */
	VAR_STRUCT  var[MAXVARS];
} VAR_TAB;


static VAR_TAB  v;                      /* variable table                       */
static MW       loc_vars[MAXVARS];      /* local variable store for decompile   */
static MW       tmp_vars[MAXTMPS];      /* temp variable store for decompile    */
static MW       clause_name;            /* name of current compiled clause      */
static MW       clause_arity;           /* arity of current compiled clause     */
static int      first_goal;             /* flag for decompile                   */
static int      is_block;               /* block instruction                    */
static CODEPTR  codeptr;                /* pointer to the last generated code   */
static jmp_buf  compiler_err;           /* error recovery                       */


static void compile_goal(MWPTR goal);
static void compile_struct(MW arity, MWPTR sptr);
static int  find_var(MWPTR var_address);
static void compute_var_offsets(void);
static void gen_head_code(MWPTR goal);
static void gen_head_struct(MW arity, MWPTR sptr);
static void gen_goal_code(MWPTR goal);
static void gen_goal_struct(MW arity, MWPTR sptr);

static void gen_nop_code(int instr);
static void gen_sop_code(int instr, int arg1);
static void gen_lop_code(int instr, MW arg1);
static void gen_sopsop_code(int instr, int arg1, int arg2);
static void gen_soplop_code(int instr, int arg1, MW arg2);

static int goal_decompile(MWPTR gptr);
static int h_struct_decompile(MWPTR gptr, MW arity);
static int g_struct_decompile(MWPTR gptr, MW arity);


/* compile clause goal at begin or end, return code in headcodeptr       */
/* sets v, clause_name, clause_arity, is_block, codeptr and compiler_err */

int compile_clause(MW goal, int atbegin, CODEPTR *headcodeptr)
{
MWPTR   gptr,           /* pointer to goal      */
        head,           /* pointer to head      */
        sptr,           /* pointer to structure */
        pptr;           /* pointer to procedure */

ErrorNumber = setjmp(compiler_err);
if (ErrorNumber)
	{
	return(ErrorNumber);
	}
v.goal_count = 0;
v.headgoal_count = 1;
v.goal_nr = 0;
v.all_vars = 0;
v.head_vars = 0;
is_block = FALSE;
gptr = &goal;
StructDeref(goal, gptr);
if (Struct(goal) && (XtrFunctor(sptr = XtrStruct(goal)) == RuleFunctor))
	{
	/* compile clause */
	gptr = FirstArgPtr(sptr);
	compile_goal(head = gptr++);
	goal = *gptr;
	ValDeref(goal);
	/* compile body */
	while (Struct(goal) && (XtrFunctor(sptr = XtrStruct(goal)) == CommaFunctor))
		{
		gptr = FirstArgPtr(sptr);
		compile_goal(gptr++);
		goal = *gptr;
		ValDeref(goal);
		}
	compile_goal(gptr);
	compute_var_offsets();
	/* generate code */
	*headcodeptr = codeptr = AddClauseHeader(heapptr);
	gen_head_code(head);
	goal = *(gptr = head + 1);
	ValDeref(goal);
	while (Struct(goal) && (XtrFunctor(sptr = XtrStruct(goal)) == CommaFunctor))
		{
		gptr = FirstArgPtr(sptr);
		gen_goal_code(gptr++);
		goal = *gptr;
		ValDeref(goal);
		}
	gen_goal_code(gptr);
	}
else
	{
	/* compile fact */
	compile_goal(gptr);
	compute_var_offsets();
	*headcodeptr = codeptr = AddClauseHeader(heapptr);
	gen_head_code(gptr);
	}
heapptr = AdjustCodePtr(codeptr);
codeptr = *headcodeptr;
FrameSize(codeptr) = v.frame_size;
pptr = find_proc(find_func(clause_name, clause_arity));
insert_clause(pptr, codeptr, atbegin);
indexing(pptr+1);
return(Success);
}


/* compile the goal at gptr            */
/* uses v, is_block and compiler_err  */

static void compile_goal(MWPTR gptr)
{
MW      goal;
MW      i;				/* var index or arity */
MWPTR   functor;        /* functor of goal    */

goal = *gptr;
StructDeref(goal, gptr);
if (RefUndef(goal))
	{
	i = find_var(gptr);
	if ((v.goal_count == 0) || (v.var[i].var_count++ == 0))
		longjmp(compiler_err, InvalidGoalError);
	v.var[i].last_goal = v.goal_count++;
	return;
	}
if (Atom(goal))
	{
	goal = XtrAtomIndex(goal);
	i = 0;
	if ((functor = find_func(goal, i)) == 0)
		longjmp(compiler_err, ErrorNumber);
	}
else if(Struct(goal))
	{
	gptr = XtrStruct(goal);
	functor = XtrFunctor(gptr);
	i = FunctArity(functor);
	goal = FunctIndex(functor);
	gptr = FirstArgPtr(gptr);
	compile_struct(i, gptr);
	}
else
	longjmp(compiler_err, InvalidGoalError);
if ((functor == CutFunctor) || (find_bip(goal, i) >= 0))
	{
	if (v.goal_count == 0)
		longjmp(compiler_err, InvalidGoalError);
	if (v.goal_count == v.headgoal_count)
		v.headgoal_count++;
	if (functor == BlockFunctor)
		{
		is_block = TRUE;
		v.goal_count++;
		}
	else if (functor == ExitBlockFunctor)
		v.goal_count++;
	}
else if (find_proc(functor) == 0)
	longjmp(compiler_err, OutOfMemoryError);
v.goal_count++;
}


/* compile struct sptr with arity arity  */

static void compile_struct(MW arity, MWPTR sptr)
{
MW      m;			/* tagged value cell */
MWPTR   mptr;		/* pointer to m      */
int     i;			/* var index         */

while (--arity >= 0)
	{
	m = *(mptr = sptr++);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		i = find_var(mptr);
		v.var[i].var_count++;
		v.var[i].last_goal = v.goal_count;
		}
	else if (List(m))
		compile_struct(2, XtrList(m));
	else if (Struct(m))
		{
		mptr = XtrStruct(m);
		compile_struct(XtrArity(mptr), FirstArgPtr(mptr));
		}
	} 
}


static int find_var(MWPTR var_address)
{
int i;

i = v.all_vars;
while (--i >= 0)
	if (var_address == v.var[i].address)
		return(i);
if (v.all_vars == MAXVARS)
	longjmp(compiler_err, OutOfVarsError);
i = v.all_vars++;
v.var[i].address = var_address;
v.var[i].type = VOID_VAR;
v.var[i].var_count = 0;
v.var[i].var_nr = 0;
v.var[i].first_goal = v.goal_count;
return(i);
}


static void compute_var_offsets(void)
{
int   i;
SHORT o;
if (v.headgoal_count == v.goal_count)
	v.frame_size = 0;
else
	v.frame_size = 2 * (SIZEMW / SIZECHAR);
if (is_block)
	v.frame_size += BlockSize * (SIZEMW / SIZECHAR);
i = v.all_vars;
while (--i >= 0)
	if (v.var[i].var_count > 1)
		if (v.var[i].last_goal < v.headgoal_count)
			v.var[i].type = TEMP_VAR;
		else
			{
			v.var[i].type = LOCAL_VAR;
			v.frame_size += (SIZEMW / SIZECHAR);
			v.var[i].offset = -v.frame_size;
			if (v.var[i].first_goal < v.headgoal_count)
				v.head_vars++;
			}
o = 0;
for (i = 0; i < v.all_vars; i++)
	if (v.var[i].type == TEMP_VAR)
		if (o < MAXTMPS) 
			{
			v.var[i].offset = o * (SIZEMW / SIZECHAR);
			o++;
			}
		else
			{
			v.var[i].type = LOCAL_VAR;
			v.frame_size += (SIZEMW / SIZECHAR);
			v.var[i].offset = -v.frame_size;
			v.head_vars++;  /* if (v.var[i].first_goal < v.headgoal_count) */
			}
}


static void gen_head_code(MWPTR gptr)
{
MW      goal;

goal = *gptr;
StructDeref(goal, gptr);
CodeStackCheck;
if (Atom(goal))
	{
	clause_name = XtrAtomIndex(goal);
	clause_arity = 0;
	}
else /* Struct(goal) */
	{
	gptr = XtrStruct(goal);
	clause_name = FunctIndex(XtrFunctor(gptr));
	clause_arity = XtrArity(gptr);
	}
if (clause_arity > 0)
	gen_head_struct(clause_arity, FirstArgPtr(gptr));
if (++v.goal_nr == v.goal_count)
	gen_nop_code(c_nogoal);
}


static void gen_head_struct(MW arity, MWPTR sptr)
{
MW      m;
MWPTR   mptr, functor;
int     i;

while (--arity >= 0)
	{
	m = *(mptr = sptr++);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		i = find_var(mptr);
		switch(v.var[i].type)
			{
			case VOID_VAR:
				gen_nop_code(h_void);
				break;
			case TEMP_VAR:
				if (v.var[i].var_nr++ == 0)
					if (v.var[i].offset == 0)
						gen_nop_code(h_fsttmp0);
					else
						gen_sop_code(h_fsttemp, (int) v.var[i].offset);
				else
					if (v.var[i].offset == 0)
						gen_nop_code(h_nxttmp0);
					else
						gen_sop_code(h_nxttemp, (int) v.var[i].offset);
				break;
			case LOCAL_VAR:
				if (v.var[i].var_nr++ == 0)
					gen_sop_code(h_fstvar, (int) v.var[i].offset);
				else
					gen_sop_code(h_nxtvar, (int) v.var[i].offset);
				break;
			}
		}
	else if (Nil(m))
		gen_nop_code(h_nil);
	else if(Int(m) || Atom(m))
		gen_lop_code(h_const, m);
	else if (List(m))
		{
		gen_nop_code(h_list);
		gen_head_struct(2, XtrList(m));
		}
	else /* Struct(m) */
		{
		mptr = XtrStruct(m);
		functor = XtrFunctor(mptr);
		gen_lop_code(h_struct, (MW) functor);
		gen_head_struct(FunctArity(functor), FirstArgPtr(mptr));
		}
	} 
}


static void gen_goal_code(MWPTR gptr)
{
MW      goal, name, arity, m;
MWPTR   pptr, mptr, functor;
int     i, inst;
int     bipnr;

goal = *gptr;
StructDeref(goal, gptr);
if (RefUndef(goal))
	{
	i = find_var(gptr);
	v.var[i].var_nr++;
	if (v.goal_nr == v.headgoal_count)
		gen_sopsop_code(c_metacall, (int) v.head_vars, g_meta);
	else
		gen_sop_code(c_metacall, g_meta);
	gen_sop_code(c_metacall, (int) v.var[i].offset);
	bipnr = -1;
	}
else
	{
	if (Atom(goal))
		{
		name = XtrAtomIndex(goal);
		arity = 0;
		functor = find_func(name, arity);
		}
	else /* Struct(goal) */
		{
		gptr = XtrStruct(goal);
		functor = XtrFunctor(gptr);
		name = FunctIndex(functor);
		arity = FunctArity(functor);
		gptr = FirstArgPtr(gptr);
		}
	if (functor == CutFunctor)
		{
		bipnr = 0;
		gen_nop_code(c_cut);
		if (v.goal_nr > v.headgoal_count)
			gen_nop_code(v.frame_size);
		}
	else if ((bipnr = find_bip(name, arity)) >= 0)
		gen_sop_code(c_builtin, (int) bipnr);
	else
		{
		pptr = FunctProc(functor);
		inst = c_goal;
		if (arity > 0)
			{
			m = *(mptr = gptr);
			StructDeref(m, mptr);
			if (RefUndef(m))
				{
				i = find_var(mptr);
				if ((v.var[i].type == LOCAL_VAR) && (v.var[i].var_nr != 0))
					{
					gptr++;
					arity--;
					v.var[i].var_nr++;
					i = v.var[i].offset;
					inst = c_xgoal;
					}
				}
			}
		if (v.goal_nr == v.headgoal_count)
			switch (v.head_vars)
				{
				case 1:
					if (inst == c_goal)
						gen_lop_code(c_1_goal, (MW) pptr);
					else
						gen_lop_code(c_1_xgoal, (MW) pptr);
					break;
				case 2:
					if (inst == c_goal)
						gen_lop_code(c_2_goal, (MW) pptr);
					else
						gen_lop_code(c_2_xgoal, (MW) pptr);
					break;
				case 3:
					if (inst == c_goal)
						gen_lop_code(c_3_goal, (MW) pptr);
					else
						gen_lop_code(c_3_xgoal, (MW) pptr);
					break;
				case 4:
					if (inst == c_goal)
						gen_lop_code(c_4_goal, (MW) pptr);
					else
						gen_lop_code(c_4_xgoal, (MW) pptr);
					break;
				default:
					gen_soplop_code(inst, (int) v.head_vars, (MW) pptr);
				}
		else
			gen_lop_code(inst, (MW) pptr);
		if (inst == c_xgoal)
			gen_nop_code(i);
		}
	if (arity > 0)
		if ((v.goal_nr < v.headgoal_count) || (bipnr >= 0))
			gen_head_struct(arity, gptr);
		else
			gen_goal_struct(arity, gptr);
	}
v.goal_nr++;
if (bipnr < 0)
	if (v.goal_nr == v.goal_count)
		gen_nop_code(c_lastcall);
	else
		gen_nop_code(c_call);
else if (functor == ExitBlockFunctor)
	gen_nop_code(c_exit_block);
else if (functor == BlockFunctor)
	{
	if (v.goal_nr++ == v.headgoal_count)
		gen_sopsop_code(c_block_in, (int) v.head_vars, g_meta);
	else
		gen_sop_code(c_block_in, g_meta);
	gen_sop_code(c_block_in, (int) -((BlockSize + 2) * (SIZEMW / SIZECHAR)));
	gen_nop_code(c_call);
	gen_sop_code(c_block_out, (int) -((BlockSize + 2) * (SIZEMW / SIZECHAR)));
	if (v.goal_nr == v.goal_count)
		gen_nop_code(c_lastcall);
	else
		gen_nop_code(c_call);
	}
else if (v.goal_nr == v.goal_count)
	gen_nop_code(c_nogoal);
}


static void gen_goal_struct(MW arity, MWPTR sptr)
{
MW      m;
MWPTR   mptr, functor;
int     i;

while (--arity >= 0)
	{
	m = *(mptr = sptr++);
	StructDeref(m, mptr);
	if (RefUndef(m))
		{
		i = find_var(mptr);
		if (v.var[i].type == VOID_VAR)
			gen_nop_code(g_void);
		else
			if (v.var[i].var_nr++ == 0)
				switch ((int) v.var[i].offset)
					{
					case Offset(0):
						gen_nop_code(g_0fstvar);
						break;
					case Offset(1):
						gen_nop_code(g_1fstvar);
						break;
					case Offset(2):
						gen_nop_code(g_2fstvar);
						break;
					case Offset(3):
						gen_nop_code(g_3fstvar);
						break;
					default:
						gen_sop_code(g_fstvar, (int) v.var[i].offset);
					}
			else
				switch ((int) v.var[i].offset)
					{
					case Offset(0):
						gen_nop_code(g_0nxtvar);
						break;
					case Offset(1):
						gen_nop_code(g_1nxtvar);
						break;
					case Offset(2):
						gen_nop_code(g_2nxtvar);
						break;
					case Offset(3):
						gen_nop_code(g_3nxtvar);
						break;
					default:
						gen_sop_code(g_nxtvar, (int) v.var[i].offset);
					}
		}
	else if (Nil(m))
		gen_nop_code(g_nil);
	else if(Int(m) || Atom(m))
		gen_lop_code(g_const, m);
	else if (List(m))
		{
		gen_nop_code(g_list);
		gen_goal_struct(2, XtrList(m));
		}
	else /* Struct(m) */
		{
		mptr = XtrStruct(m);
		functor = XtrFunctor(mptr);
		gen_lop_code(g_struct, (MW) functor);
		gen_goal_struct(FunctArity(functor), FirstArgPtr(mptr));
		}
	} 
}


static void gen_nop_code(int instr)
{
InstrStore(codeptr, instr);
CodeStackCheck;
}


static void gen_sop_code(int instr, int arg1)
{
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
CodeStackCheck;
}


static void gen_lop_code(int instr, MW arg1)
{
InstrStore(codeptr, instr);
LopStore(codeptr, arg1);
CodeStackCheck;
}


static void gen_sopsop_code(int instr, int arg1, int arg2)
{
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
SopStore(codeptr, arg2);
CodeStackCheck;
}


static void gen_soplop_code(int instr, int arg1, MW arg2)
{
InstrStore(codeptr, instr);
SopStore(codeptr, arg1);
LopStore(codeptr, arg2);
CodeStackCheck;
}


int clause_decompile(MWPTR functor, CODEPTR clauseptr, MWPTR termptr)
{
MW      m;
MWPTR   mptr;
int     err;

ErrorNumber = setjmp(compiler_err);
if (ErrorNumber)
	{
	return(ErrorNumber);
	}
codeptr = clauseptr;
mptr = termptr;
if (FunctArity(functor) == 0)
	*termptr = MakeAtom(FunctIndex(functor));
else
	{
	CreateStruct(termptr, functor, FunctArity(functor));
	if ((err = h_struct_decompile(termptr, FunctArity(functor))) != Success)
		return(err);
	}
if (InstrRead(codeptr) == c_nogoal)
	return(Success);
first_goal = TRUE;
m = *mptr;
CreateStruct(mptr, RuleFunctor, 2);
*mptr++ = m;
while (goal_decompile(mptr))
	{
	m = *mptr;
	CreateStruct(mptr, CommaFunctor, 2);
	*mptr++ = m;
	}
return(Success);
}


static int goal_decompile(MWPTR gptr)
{
MW      m, arity;
MWPTR   pptr, functor;
int     i, err;

switch (InstrFetch(codeptr))
	{
	case c_1_goal:
	case c_2_goal:
	case c_3_goal:
	case c_4_goal:
		goto goal_decompile_0;
	case c_goal:
		if (first_goal)
			{
			SIncr(codeptr);
goal_decompile_0:
			first_goal = FALSE;
			}
		LopFetch(m, codeptr);
		pptr = (MWPTR) m;
		functor = ProcFunctor(pptr);
		if (FunctArity(functor) == 0)
			*gptr = MakeAtom(FunctIndex(functor));
		else
			{
			CreateStruct(gptr, functor, FunctArity(functor));
			if ((err = g_struct_decompile(gptr, FunctArity(functor))) != Success)
				return(err);
			}
		if (InstrFetch(codeptr) == c_call)
			return(TRUE);
		return(FALSE);
	case c_1_xgoal:
	case c_2_xgoal:
	case c_3_xgoal:
	case c_4_xgoal:
		goto goal_decompile_1;
	case c_xgoal:
		if (first_goal)
			{
			SIncr(codeptr);
goal_decompile_1:
			first_goal = FALSE;
			}
		LopFetch(m, codeptr);
		pptr = (MWPTR) m;
		functor = ProcFunctor(pptr);
		CreateStruct(gptr, functor, FunctArity(functor));
		*gptr++ = Var(codeptr);
		if ((err = g_struct_decompile(gptr, FunctArity(functor) - 1)) != Success)
			return(err);
		if (InstrFetch(codeptr) == c_call)
			return(TRUE);
		return(FALSE);
	case c_cut:
		*gptr = MakeAtom(CutAtomIndex);
		if (! first_goal)
			SIncr(codeptr);
		if (InstrRead(codeptr) != c_nogoal)
			return(TRUE);
		return(FALSE);
	case c_builtin:
		i = SopFetch(codeptr);
		m = builtin[i].bt_name;
		arity = builtin[i].bt_arity;
		if (arity == 0)
			*gptr = MakeAtom(m);
		else
			{
			CreateStruct(gptr, find_func(m, arity), arity);
			if ((err = h_struct_decompile(gptr, arity)) != Success)
				return(err);
			}
		switch (InstrRead(codeptr))
			{
			case c_block_in:
				SIncr(codeptr);             /* c_block_in       */
				if (first_goal)
					{
					first_goal = FALSE;
					SIncr(codeptr);
					}
				SIncr(codeptr);             /* g_meta           */
				SIncr(codeptr);             /* c_block_in       */
				SIncr(codeptr);             /* var_offset       */
				SIncr(codeptr);             /* c_call           */
				SIncr(codeptr);             /* c_block_out      */
				SIncr(codeptr);             /* var_offset       */
				if (InstrFetch(codeptr) == c_call)
					return(TRUE);
				/* fall through to c_exit_block */
			case c_nogoal:
				return(FALSE);
			default:
				return(TRUE);
			}
	case c_metacall:
		if (first_goal)
			{
			first_goal = FALSE;
			SIncr(codeptr);
			}
		SIncr(codeptr);
		SIncr(codeptr);
		*gptr = Var(codeptr);
		if (InstrFetch(codeptr) == c_call)
			return(TRUE);
		return(FALSE);
	case c_exit_block:
		*gptr = MakeAtom(ExitBlockAtomIndex);
		return(FALSE);
	default:
		longjmp(compiler_err, SystemError);
		return(FALSE);
	}
}


static int h_struct_decompile(MWPTR gptr, MW arity)
{
MW      m;
MWPTR   mptr;
MW      i;
int     err;

while (--arity >= 0)
	switch (HeadMux(InstrFetch(codeptr)))
		{
		case HeadMux(h_const):
			LopFetch(m, codeptr);
			*gptr++ = m;
			break;
		case HeadMux(h_nil):
			*gptr++ = MakeNil;
			break;
		case HeadMux(h_list):
			*gptr++ = MakeList(copyptr -= 2);
			if ((err = h_struct_decompile(copyptr, 2)) != Success)
				return(err);
			break;
		case HeadMux(h_struct):
			mptr = gptr++;
			FetchStruct(mptr, codeptr, i = ArityFetch(codeptr));
			if ((err = h_struct_decompile(mptr, i)) != Success)
				return(err);
			break;
		case HeadMux(h_fstvar):
			Var(codeptr) = MakeRef(gptr);
		case HeadMux(h_void):
			*gptr++ = MakeUndef(gptr);
			break;
		case HeadMux(h_nxtvar):
			*gptr++ = Var(codeptr);
			break;
		case HeadMux(h_fsttemp):
			Tmp(codeptr) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case HeadMux(h_nxttemp):
			*gptr++ = Tmp(codeptr);
			break;
		case HeadMux(h_fsttmp0):
			ShortTmp(0) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case HeadMux(h_nxttmp0):
			*gptr++ = ShortTmp(0);
			break;
		default:
			longjmp(compiler_err, SystemError);
		}
return(Success);
}


static int g_struct_decompile(MWPTR gptr, MW arity)
{
MW      m;
MWPTR   mptr;
MW      i;
int     err;

while (--arity >= 0)
	switch (GoalMux(InstrFetch(codeptr)))
		{
		case GoalMux(g_const):
			LopFetch(m, codeptr);
			*gptr++ = m;
			break;
		case GoalMux(g_nil):
			*gptr++ = MakeNil;
			break;
		case GoalMux(g_list):
			*gptr++ = MakeList(copyptr -= 2);
			if ((err = g_struct_decompile(copyptr, 2)) != Success)
				return(err);
			break;
		case GoalMux(g_struct):
			mptr = gptr++;
			FetchStruct(mptr, codeptr, i = ArityFetch(codeptr));
			if ((err = g_struct_decompile(mptr, i)) != Success)
				return(err);
			break;
		case GoalMux(g_fstvar):
			Var(codeptr) = MakeRef(gptr);
		case GoalMux(g_void):
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_nxtvar):
			*gptr++ = Var(codeptr);
			break;
		case GoalMux(g_0fstvar):
			ShortVar(0) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_0nxtvar):
			*gptr++ = ShortVar(0);
			break;
		case GoalMux(g_1fstvar):
			ShortVar(1) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_1nxtvar):
			*gptr++ = ShortVar(1);
			break;
		case GoalMux(g_2fstvar):
			ShortVar(2) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_2nxtvar):
			*gptr++ = ShortVar(2);
			break;
		case GoalMux(g_3fstvar):
			ShortVar(3) = MakeRef(gptr);
			*gptr++ = MakeUndef(gptr);
			break;
		case GoalMux(g_3nxtvar):
			*gptr++ = ShortVar(3);
			break;
		default:
			longjmp(compiler_err, SystemError);
		}
return(Success);
}


int b_asserta(MW arg[])
{
CODEPTR codeptr;
return(compile_clause(arg[0], TRUE, &codeptr));
}


int b_assertz(MW arg[])
{
CODEPTR codeptr;
return(compile_clause(arg[0], FALSE, &codeptr));
}
