#include "mini.h"
#include "time.h" 
#include <sys/timeb.h>

int     ErrorNumber, i;
FILE    *infile;
FILE    *outfile;
MW      infile_name;
MW      outfile_name;

void m_error(int errnr);

static CODEPTR firstclauseptr;

static int compile(MW *fname);

FILE *zfile;

/* calls the compiler and the VAM2&1P-Interpreter */
int main(int argc, char *argv[])
{
MW      fname;
size_t  heap_copy_size, trail_stack_size;

#ifdef zwischencodeausgabe
	if ((zfile = fopen(zfilename,"a")) == NULL)
		printf("cannot open file\n");
#endif

heap_copy_size = HeapCopySize;
trail_stack_size = TrailStackSize;
if (mm_initialise(heap_copy_size, trail_stack_size))
	{
	m_error(ErrorNumber);
	exit(1);
	}
firstclauseptr = 0;

infile = stdin;
outfile = stdout;
fname = infile_name = outfile_name = UserAtom;

/* clauses per file */
if (argc > 1)
	(void) find_atom(argv[1], &fname);
else
/* clauses input manual */
	printf("enter clauses, argumentless start clause first, end with ^Z\n\n");

/* call compiler */
(void) compile(&fname);

/* call interpreter */
m_error(mi_interpreter(firstclauseptr));

#ifdef Atari
(void) getchar();
#endif

#ifdef zwischencodeausgabe
  fclose(zfile);
#endif

return(Success);
}


/* compiles clause after clause from file or stdin (fname) */
static int compile(MW *fname)
{
MWPTR   copysave;
MW      term;
CODEPTR clauseptr;
int     err;

if (b_see(fname) != Success)
	return(ErrorNumber = FileOpenError);

copysave = copyptr;                                                                  /* stackchange: no modification */
for (;;)
	{
	copyptr = copysave;                                                          /* stackchange: no modification */
	if ((b_eof() != Fail) || ((err = scan_term(&term)) == EndOfInputError))
		{
		(void) b_seen();
		copyptr = copysave;     /* stackchange: no modification */
		return(Success);
		}
	if (err == Success)
		{
		(void) compile_clause(term, FALSE, &clauseptr);
		if (firstclauseptr == 0)
			firstclauseptr = clauseptr;
		}
	}
}


/* error handling routine */
void m_error(int errnr)
{
switch(errnr)
	{
	case Success:
		printf("No Error, Success\n");
		break;
	case Fail:
		printf("No Error, Fail\n");
		break;
	case SystemError:
		printf("System Error\n");
		break;
	case SyntaxError:
		printf("Syntax Error\n");
		break;
	case OperatorError:
		printf("Operator Precedence Error\n");
		break;
	case EndOfInputError:
		printf("Error: Unexpected End Of Input\n");
		break;
	case OutOfMemoryError:
		printf("Error: Out Of Memory\n");
		break;
	case FileOpenError:
		printf("Error Opening File\n");
		break;
	case UndefProcError:
		printf("Error: Calling undefined procedure\n");
		break;
	case NotImplementedError:
		printf("SystemError: Not yet implemented\n");
		break;
	case InvalidGoalError:
		printf("Error: Asserting invalid goal\n");
		break;
	case OutOfVarsError:
		printf("Error: Too many variables\n");
		break;
	case TooManyFilesError:
		printf("Error: Too many active files\n");
		break;
	case ExitTrueError:
		printf("Exit with success\n");
		break;
	case ExitFailError:
		printf("Exit with fail\n");
		break;
	case BuiltinInstError:
		printf("Wrong instantiated builtin argument\n");
		break;
	case BuiltinTypeError:
		printf("Wrong type of builtin argument\n");
		break;
	case BuiltinRangeError:
		printf("Builtin argument out of range\n");
		break;
	case BuiltinOpError:
		printf("Invalid operator definition\n");
		break;
	case StackOverflowError:
		printf("Overflow of environment stack\n");
		break;
	case CopyOverflowError:
		printf("Overflow of copy stack / trail stack\n");
		break;
	case NoBlockError:
		printf("Error: No matching block for exit_block\n");
		break;
	default:
		printf("Unknown Error\n");
	}
}

