#include "mini.h"

/* character classes */

#define UC       1      /* upper case alphabetic    */
#define UL       2      /* underline character      */
#define LC       3      /* lower case alphabetic    */
#define DI       4      /* digit                    */
#define SQ       5      /* single quote             */
#define DQ       6      /* double quote             */
#define SY       7      /* symbol character         */
#define FS       8      /* fullstop                 */
#define SL       9      /* solo character           */
#define LP      10      /* left paranthesis         */
#define RP      11      /* right paranthesis        */
#define LB      12      /* left bracket             */
#define RB      13      /* right bracket            */
#define LCB     14      /* left curly bracket       */
#define RCB     15      /* right curly bracket      */
#define CO      16      /* comma                    */
#define VB      17      /* vertical bar             */
#define BS      18      /* blank space              */
#define NL      19      /* new line                 */
#define CM      20      /* comment                  */
#define C1      21      /* comment char 1 /         */
#define C2      22      /* comment char 2 *         */
#define C3      23      /* combined comment char 12 */
#define EM      24      /* end of medium            */

static BYTE cclass[256] = {
/* nul soh stx etx eot enq ack bel bs  ht  nl  vt  np  cr  so  si  */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, NL, BS, BS, NL, BS, BS,

/* dle dc1 dc2 dc3 dc4 nak syn etb can em  sub esc fs  gs  rs  us  */
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, EM, BS, BS, BS, BS, BS,

/* sp  !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /   */
   BS, SL, DQ, SY, SY, CM, SY, SQ, LP, RP, C2, SY, CO, SY, FS, C1,

/* 0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?   */
   DI, DI, DI, DI, DI, DI, DI, DI, DI, DI, SY, SL, SY, SY, SY, SY,

/* @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O   */
   SY, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC,

/* P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _   */
   UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, UC, LB, SY, RB, SY, UL,

/* `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o   */
   SY, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC,

/* p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~   del */
   LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LC, LCB,VB,RCB, SY, BS,

   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
   BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS, BS,
};

static jmp_buf  io_err;

static BYTE     cc;                 /* character class (see above)      */
static int      cread;              /* character read                   */
static int      cprer = 0 ;         /* character preread                */
static int      readmode = FALSE;   /* readmode = TRUE, getmode = FALSE */
static MWPTR    varbase;            /* root of variable list            */
static MWPTR    varlist;            /* ptr to variable space on trail   */

#define SetNoPreRead        cprer = 0
#define SetCharPreRead      cprer = 1
#define SetComPreRead       cprer = -1
#define CharPreRead         cprer
#define NotCharPreRead      (cprer == 0)
#define ComPreRead          (cprer < 0)

#define NilToken             0
#define IntToken             1
#define RealToken            2
#define VarToken             3
#define AtomToken            4
#define FuncToken            5
#define LParToken            6
#define RParToken            7
#define CommaToken           8
#define LBrackToken          9
#define RBrackToken         10
#define LCurlyBrackToken    11
#define RCurlyBrackToken    12
#define VBarToken           13
#define StringToken         14
#define FullStopToken       15
#define EofToken            16

#define EmptySet             0
#define CommaSet             1
#define RParSet              2
#define VBarSet              4
#define RBrackSet            8
#define RCurlyBrackSet      16
#define FStopSet            32

#define MaxFile             16
#define WriteDepth          15

#define In(setel, set)      ((setel)&(set))
#define AdjustPtr(mp, cp)   (mp)=((MWPTR)((((long)cp)+3)&((long)-4)))
#define NextChar()          if((cread=getc(infile))==EOF)cread=26;cc=cclass[cread]

#define NotNextToken        ((ErrorNumber=next_token())!=Success)
#define ErrorAtomIndex      (ATOM_MASK-1)
#define ErrorAtom           MakeAtom(ErrorAtomIndex)

MWPTR   ErrorFunctor1, ErrorFunctor2, NullFunctor1, NullFunctor2, CurlyFunctor2;

typedef struct FILE_TAB {
	MW      f_name;
	FILE    *f_file;
	int     f_is_infile;
	int     f_char;
	int     f_cprer;
	int     f_readmode;
	} FILE_TAB;

static FILE_TAB filetab[MaxFile];
static usedfiles = 0;

static int token, settoken;
static MW  tokenval;

static int change_file(MW name, int is_infile);
static int close_file(int is_infile);
static int next_token(void);
static int scan_list(MWPTR mptr);
static int scan_args(MW *arityptr);
static int scan_subterm(MWPTR mptr, int prec, int follow);
static void write_atom(CHARPTR atomptr, int quoted);


/* closes the current output/input file and opens the file name as the current output/input file */
static int change_file(MW name, int is_infile)
{
MW      c_name;
FILE    *c_file;
int     c_char;
int     c_cprer;
int     c_readmode;
int     i;

if (is_infile)
	{
	if (name == infile_name)
		return(Success);
	else if ((name == outfile_name) && (name != UserAtom))
		return(Fail);
	}
else if (name == outfile_name)
	return(Success);
else if ((name == infile_name) && (name != UserAtom))
	return(Fail);
i = usedfiles;
while ((--i >= 0) && ((filetab[i].f_name != name) ||
	((name == UserAtom) && (filetab[i].f_is_infile != is_infile))));
if (i >= 0)
	{
	if (filetab[i].f_is_infile == is_infile)
		{
		if (is_infile)
			{
			c_name = infile_name;
			c_file = infile;
			c_char = cread;
			c_cprer = cprer;
			c_readmode = readmode;
			infile_name = name;
			infile = filetab[i].f_file;
			cread = filetab[i].f_char;
			cprer = filetab[i].f_cprer;
			readmode = filetab[i].f_readmode;
			cc = cclass[cread];
			filetab[i].f_char = c_char;
			filetab[i].f_cprer = c_cprer;
			filetab[i].f_readmode = c_readmode;
			}
		else
			{
			c_name = outfile_name;
			c_file = outfile;
			outfile_name = name;
			outfile = filetab[i].f_file;
			}
		filetab[i].f_name = c_name;
		filetab[i].f_file = c_file;
		return(Success);
		}
	fclose(filetab[i].f_file);
	if (--usedfiles > i)
		{
		filetab[i].f_name = filetab[usedfiles].f_name;
		filetab[i].f_file = filetab[usedfiles].f_file;
		filetab[i].f_is_infile = filetab[usedfiles].f_is_infile;
		filetab[i].f_char = filetab[usedfiles].f_char;
		filetab[i].f_cprer = filetab[usedfiles].f_cprer;
		filetab[i].f_readmode = filetab[usedfiles].f_readmode;
		}
	}
if (usedfiles == MaxFile)
	return(ErrorNumber = TooManyFilesError);
filetab[usedfiles].f_is_infile = is_infile;
if (is_infile)
	{
	filetab[usedfiles].f_name = infile_name;
	filetab[usedfiles].f_file = infile;
	filetab[usedfiles].f_char = cread;
	filetab[usedfiles].f_cprer = cprer;
	filetab[usedfiles].f_readmode = readmode;
	if ((c_file = open_r(XtrAtomPtr(name))) == NULL)
		return(ErrorNumber = FileOpenError);
	infile = c_file;
	infile_name = name;
	SetNoPreRead;
	readmode = FALSE;
	}
else
	{
	filetab[usedfiles].f_name = outfile_name;
	filetab[usedfiles].f_file = outfile;
	if ((c_file = open_w(XtrAtomPtr(name))) == NULL)
		return(ErrorNumber = FileOpenError);
	outfile = c_file;
	outfile_name = name;
	}
usedfiles++;
return(Success);
}


/* close file */
static int close_file(int is_infile)
{
MW      c_name;
FILE    *c_file;
int     i;

if (is_infile)
	{
	c_file = infile;
	c_name = infile_name;
	}
else
	{
	c_file = outfile;
	c_name = outfile_name;
	}
if (c_name == UserAtom)
	{
	if (is_infile)
		{
		SetNoPreRead;
		readmode = FALSE;
		}
	return(Success);
	}
fclose(c_file);
i = usedfiles;
while ((--i >= 0) && ((filetab[i].f_name != UserAtom) ||
	(filetab[i].f_is_infile != is_infile)));
if (is_infile)
	{
	infile_name = UserAtom;
	infile = filetab[i].f_file;
	cread = filetab[i].f_char;
	cprer = filetab[i].f_cprer;
	readmode = filetab[i].f_readmode;
	cc = cclass[cread];
	}
else
	{
	outfile_name = UserAtom;
	outfile = filetab[i].f_file;
	}
if (--usedfiles > i)
	{
	filetab[i].f_name = filetab[usedfiles].f_name;
	filetab[i].f_file = filetab[usedfiles].f_file;
	filetab[i].f_is_infile = filetab[usedfiles].f_is_infile;
	filetab[i].f_char = filetab[usedfiles].f_char;
	filetab[i].f_cprer = filetab[usedfiles].f_cprer;
	filetab[i].f_readmode = filetab[usedfiles].f_readmode;
	}
return(Success);
}


/* reads and evaluates next token */
static int next_token(void)
{
char    *cptr, string[2];
MW      m;
MWPTR   vlistptr;

string[1] = 0;
settoken = EmptySet;
if (ComPreRead)
	{
	SetCharPreRead;
	goto c1_read;
	}

next_token_start:

switch((int) cc)
	{
	case UC:                                  /* upper case alphabetic    */
	case UL:                                  /* underline character      */
		cptr = (char *) (varlist+2);
		do  {
			*cptr++ = cread;
			NextChar();
			} while (cc <= DI);
		*cptr++ = 0;
		token = VarToken;
		tokenval = MakeVarUndef;
		if (strcmp((char *)(varlist+2), "_") == 0)
			return(Success);
		vlistptr = varbase;
		while (vlistptr != NULL)
			if (strcmp((char *)(varlist+2), (char *)(vlistptr+2)) == 0)
				{
				if (VarUndef(*(vlistptr+1)))
					tokenval = MakeRef(vlistptr+1);
				else
					tokenval = *(vlistptr+1);
				return(Success);
				}
			else
				vlistptr = (MWPTR)(*vlistptr);
		*varlist = (MW)varbase;
		varbase = varlist;
		*(varlist+1) = MakeVarUndef;
		tokenval = MakeRef(varlist+1);
		AdjustPtr(varlist, cptr);
		return(Success);
	case LC:				  /* lower case alphabetic    */
		cptr = (char *) varlist;
		do  {
			*cptr++ = cread;
			NextChar();
			} while (cc <= DI);
next_token_atom:
		*cptr++ = 0;
		if (cc == LP)
			{
			token = FuncToken;
			NextChar();
			}
		else
			token = AtomToken;
		return(find_atom((char*)varlist, &tokenval));
	case DI:				  /* digit		    */
		tokenval = 0;
		do  {
			tokenval = tokenval * 10 + (cread - '0');
			NextChar();
			} while (cc == DI);
		token = IntToken;
		tokenval = MakeInt(tokenval);
		return(Success);
	case SQ:				  /* single quote	     */
		cptr = (char *) varlist;
		for (;;)
			{
			NextChar();
			if (cc == SQ)
				{
				NextChar();
				if (cc != SQ)
					goto next_token_atom;	   
				}
			*cptr++ = cread;
			}
	case DQ:				  /* double quote	     */
		tokenval = MakeList(copyptr - 2);      /* stackchange: no modification */
		for (;;)
			{
			NextChar();
			if (cc == DQ)
				{
				NextChar();
				if (cc != DQ)
					{
					*(copyptr+1) = MakeNil;   /* stackchange: no modification */
					token = StringToken;
					return(Success);
					}
				}
			string[0] = cread;
			m = MakeList(copyptr-4);		   /* stackchange: no modification */
			PushCopyStack(m);
			CallAndCheck(find_atom(string, &m))
			PushCopyStack(m);
			}
	case SY:				  /* symbol character	 */
	case C2:				  /* comment char 2 *	 */
		cptr = (char*) varlist;
		*cptr++ = cread;
		NextChar();
next_token_symbol:
		for (;;)
			{
			if (cc == C1)
				{
				NextChar();
				if (cc == C2)
					{
					cc = C3;
					goto next_token_atom;	   
					}
				*cptr++ = '/';
				}
			if ((cc != FS) && (cc != SY) && (cc != C1) && (cc != C2))
				goto next_token_atom;	   
			*cptr++ = cread;
			NextChar();
			}
	case FS:				  /* fullstop		 */
		token = FullStopToken;
		NextChar();
		if ((cc == BS) || (cc == NL) || (cc == EM))
			{
			settoken = FStopSet;
			return(Success);
			}
		cptr = (char*) varlist;
		*cptr++ = '.';
		if ((cc == FS) || (cc == SY) || (cc == C1) || (cc == C2))
			goto next_token_symbol;
		goto next_token_atom;
	case SL:				  /* solo character	   */
		string[0] = cread;
		token = AtomToken;
		NextChar();
		return(find_atom(string, &tokenval));
	case LP:				  /* left paranthesis	 */
		token = LParToken;
		NextChar();
		return(Success);
	case RP:				  /* right paranthesis	*/
		token = RParToken;
		settoken = RParSet;
		NextChar();
		return(Success);
	case LB:				  /* left bracket	     */
		NextChar();
		if (cc == RB)
			{
			token = NilToken;
			tokenval = MakeNil;
			NextChar();
			}
		else
			token = LBrackToken;
		return(Success);
	case RB:				  /* right bracket	    */
		token = RBrackToken;
		settoken = RBrackSet;
		NextChar();
		return(Success);
	case LCB:				 /* left curly bracket       */
		NextChar();
		if (cc == RCB)
			{
			token = AtomToken;
			tokenval = CurlyBrackAtom;
			NextChar();
			}
		else
			token = LCurlyBrackToken;
		return(Success);
	case RCB:				 /* right curly bracket      */
		token = RCurlyBrackToken;
		settoken = RCurlyBrackSet;
		NextChar();
		return(Success);
	case CO:				  /* comma		    */
		token = CommaToken;
		tokenval = CommaAtom;
		settoken = CommaSet;
		NextChar();
		return(Success);
	case VB:				  /* vertical bar	     */
		token = VBarToken;
		tokenval = VBarAtom;
		settoken = VBarSet;
		NextChar();
		return(Success);
	case BS:				  /* blank space	      */
	case NL:				  /* new line		 */
		NextChar();
		goto next_token_start;
	case CM:				  /* comment		  */
		for (;;)
			{
			NextChar();
			if (cc == EM)
				{
				token = EofToken;
				return(Success);
				}
			if (cc == NL)
				{
				NextChar();
				goto next_token_start;
				}
			}
	case C1:				  /* comment char 1 /	 */
		NextChar();
c1_read:
		if (cc != C2)
			{
			cptr = (char *) varlist;
			*cptr++ = '/';
			goto next_token_symbol;
			}
	case C3:
		for (;;)
			{
			NextChar();
			while (cc == C2)
				{
				NextChar();
				if (cc == C1)
					{
					NextChar();
					goto next_token_start;
					}	       
				}
			if (cc == EM)
				return(ErrorNumber = EndOfInputError);
			}
	case EM:				  /* end of medium	    */
		token = EofToken;
		return(Success);
	default:
		return(ErrorNumber = SystemError);
	}
}


/* search for errors in lists */
static int scan_list(MWPTR mptr)
{
MW      m;
MWPTR   mp;
int     ErrorNumber;

if ((ErrorNumber = scan_subterm(&m, MaxPriority, CommaSet|VBarSet|RBrackSet)) == Success)
	if (token == CommaToken)
		if (NotNextToken)
			*mptr = ErrorAtom;
		else
			ErrorNumber = scan_list(mptr);
	else if (token == VBarToken)
		if (NotNextToken)
			*mptr = ErrorAtom;
		else
			{
			if ((ErrorNumber = scan_subterm(mptr, MaxPriority, RBrackSet)) == Success)
				if (NotNextToken)
					{
					PushCopyStackVar(*mptr, mp);
					PushFunctor(ErrorFunctor1);
					*mptr = MakeStruct(copyptr);	 /* stackchange: no modification */
					}
			}
	else if (token == RBrackToken)
		if (NotNextToken)
			*mptr = ErrorAtom;
		else
			*mptr = MakeNil;
	else
		{
		*mptr = ErrorAtom;
		ErrorNumber = SyntaxError;
		}
else
	*mptr = ErrorAtom;
PushCopyStackVar(*mptr, mp);
PushCopyStackVar(m, mp);
*mptr = MakeList(copyptr);               /* stackchange: no modification */
return(ErrorNumber);
}


/* search for errors in the arguments of compound terms */
static int scan_args(MW *arityptr)
{
MW      m;
MWPTR   mp;
int     ErrorNumber;

if ((ErrorNumber = scan_subterm(&m, MaxPriority, CommaSet|RParSet)) != Success)
	*arityptr = 1;
else if (token == CommaToken)
	if (NotNextToken)
		{
		PushCopyStack(ErrorAtom);
		*arityptr = 2;
		}
	else
		{
		ErrorNumber = scan_args(arityptr);
		(*arityptr)++;
		}
else if (token == RParToken)
	if (NotNextToken)
		{
		PushCopyStack(ErrorAtom);
		*arityptr = 2;
		}
	else
		*arityptr = 1;
else
	{
	m = ErrorAtom;
	ErrorNumber = SyntaxError;
	}
PushCopyStackVar(m, mp);
return(ErrorNumber);
}


/* search for errors in term mptr */
static int scan_subterm(MWPTR mptr, int prec, int follow)
{
MW      arity;
int     oprec, lprec, rprec, prec1, ErrorNumber;
MW      m, op;
MWPTR   mp, functor;

prec1 = 0;
switch (token)
	{
	case IntToken:
	case RealToken:
	case VarToken:
	case StringToken:
		*mptr = tokenval;
		if (NotNextToken)
			{
			PushCopyStackVar(*mptr, mp);
			PushFunctor(ErrorFunctor1);
			*mptr = MakeStruct(copyptr);                 /* stackchange: no modification */
			return(ErrorNumber);
			}
		break;
	case NilToken:
	case CommaToken:
	case VBarToken:
	case AtomToken:
		*mptr = tokenval;
		if (NotNextToken)
			{
			PushCopyStackVar(*mptr, mp);
			PushFunctor(ErrorFunctor1);
			*mptr = MakeStruct(copyptr);		 /* stackchange: no modification */
			return(ErrorNumber);
			}
		if (In(settoken, follow))
			return(Success);
		op = XtrAtomIndex(*mptr);
		if (!is_prefix_op(op, &oprec, &rprec))
			if (token == LParToken)
				goto scan_subterm_lpar;
			else
				break;
		if (oprec > prec)
			{
			*mptr = ErrorAtom;
			return(ErrorNumber = OperatorError);
			}
		ErrorNumber = scan_subterm(&m, rprec, follow);
		if (Int(m) && (op == MinusAtomIndex))
			{
			*mptr = MakeInt(-(XtrInt(m)));
			break;
			}
		PushCopyStackVar(m, mp);
		if ((functor = find_func(op, (MW) 1)) == 0)
			PushFunctor(NullFunctor1);
		else
			PushFunctor(functor);
		*mptr = MakeStruct(copyptr);		       /* stackchange: no modification */
		if (ErrorNumber != Success)
			return(ErrorNumber);
		prec1 = oprec;
		break;
	case FuncToken:
		*mptr = tokenval;
scan_subterm_lpar:
		if (NotNextToken)
			{
			PushCopyStack(ErrorAtom);
			if ((functor = find_func(XtrAtomIndex(*mptr), (MW) 1)) == 0)
				PushFunctor(NullFunctor1);
			else
				PushFunctor(functor);
			*mptr = MakeStruct(copyptr);		/* stackchange: no modification */
			return(ErrorNumber);
			}
		else
			ErrorNumber = scan_args(&arity);
		if ((functor = find_func(XtrAtomIndex(*mptr), arity)) == 0)
			PushFunctor(NullFunctor1);
		else
			PushFunctor(functor);
		*mptr = MakeStruct(copyptr);		      /* stackchange: no modification */
		if (ErrorNumber != Success)
			return(ErrorNumber);
		break;
	case LParToken:
		if (NotNextToken)
			{
			PushCopyStack(ErrorAtom);
			PushFunctor(NullFunctor1);
			*mptr = MakeStruct(copyptr);	       /* stackchange: no modification */
			return(ErrorNumber);
			}
		if ((scan_subterm(mptr, MaxPriority, RParSet) != Success) ||
			(token != RParToken))
			{
			PushCopyStack(ErrorAtom);
			PushCopyStackVar(*mptr, mp);
			PushFunctor(NullFunctor2);
			*mptr = MakeStruct(copyptr);	     /* stackchange: no modification */
			return(ErrorNumber = SyntaxError);
			}
		if (NotNextToken)
			{
			PushCopyStackVar(*mptr, mp);
			PushFunctor(ErrorFunctor2);
			*mptr = MakeStruct(copyptr);	     /* stackchange: no modification */
			return(ErrorNumber);
			}
		break;
	case LCurlyBrackToken:
		if (NotNextToken)
			{
			PushCopyStack(ErrorAtom);
			PushFunctor(CurlyBrackFunctor);
			*mptr = MakeStruct(copyptr);	    /* stackchange: no modification */
			return(ErrorNumber);
			}
		if ((scan_subterm(mptr, MaxPriority, RCurlyBrackSet) != Success) ||
			(token != RCurlyBrackToken))
			{
			PushCopyStack(ErrorAtom);
			PushCopyStackVar(*mptr, mp);
			PushFunctor(CurlyFunctor2);
			*mptr = MakeStruct(copyptr);	     /* stackchange: no modification */
			return(ErrorNumber = SyntaxError);
			}
		if (NotNextToken)
			{
			PushCopyStack(ErrorAtom);
			PushCopyStackVar(*mptr, mp);
			PushFunctor(CurlyFunctor2);
			*mptr = MakeStruct(copyptr);	     /* stackchange: no modification */
			return(ErrorNumber);
			}
		PushCopyStackVar(*mptr, mp);
		PushFunctor(CurlyBrackFunctor);
		*mptr = MakeStruct(copyptr);		     /* stackchange: no modification */
		break;
	case LBrackToken:
		if (NotNextToken)
			{
			PushCopyStack(ErrorAtom);
			*mptr = MakeList(copyptr);	      /* stackchange: no modification */
			return(ErrorNumber);
			}
		CallAndCheck(scan_list(mptr))
		break;
	case EofToken:
		*mptr = ErrorAtom;
		return(ErrorNumber = EndOfInputError);
	default:
		*mptr = ErrorAtom;
		return(ErrorNumber = SyntaxError);
	}

scan_subterm_postinfix:

if (In(settoken, follow))
	return(Success);
if (((token != AtomToken) && (token != CommaToken) && (token != VBarToken) &&
    (token != FuncToken)) ||
    ((token == FuncToken) &&
     (! is_infix_op(op = XtrAtomIndex(tokenval), &lprec, &oprec, &rprec) ||
     (oprec > prec) || (prec1 > lprec))))
	{
	PushCopyStackVar(*mptr, mp);
	PushFunctor(ErrorFunctor1);
	*mptr = MakeStruct(copyptr);			 /* stackchange: no modification */
	return(ErrorNumber = SyntaxError);
	}
if (token == FuncToken)
	{
	if (NotNextToken)
		{
		PushCopyStack(ErrorAtom);
		PushFunctor(NullFunctor1);
		m = MakeStruct(copyptr);                     /* stackchange: no modification */
		PushCopyStack(m);
		PushCopyStackVar(*mptr, mp);
		if ((functor = find_func(op, (MW) 2)) == 0)
			PushFunctor(NullFunctor2);
		else
			PushFunctor(functor);
		*mptr = MakeStruct(copyptr);		 /* stackchange: no modification */
		return(ErrorNumber = SyntaxError);
		}
	if ((scan_subterm(&m, MaxPriority, RParSet) != Success) ||
          (token != RParToken))
		{
		PushCopyStack(ErrorAtom);
		PushCopyStackVar(m, mp);
		PushFunctor(NullFunctor2);
		m = MakeStruct(copyptr);		    /* stackchange: no modification */
		PushCopyStack(m);
		PushCopyStackVar(*mptr, mp);
		if ((functor= find_func(op, (MW) 2)) == 0)
			PushFunctor(NullFunctor2);
		else
			PushFunctor(functor);
		*mptr = MakeStruct(copyptr);		/* stackchange: no modification */
		return(ErrorNumber = SyntaxError);
		}
	PushCopyStackVar(m, mp);
	PushCopyStackVar(*mptr, mp);
	if ((functor= find_func(op, (MW) 2)) == 0)
		PushFunctor(NullFunctor2);
	else
		PushFunctor(functor);
	*mptr = MakeStruct(copyptr);		       /* stackchange: no modification */
	if (NotNextToken)
		{
		PushCopyStack(*mptr);
		PushFunctor(ErrorFunctor1);
		*mptr = MakeStruct(copyptr);	       /* stackchange: no modification */
		return(ErrorNumber);
		}
	prec1 = oprec;
	goto scan_subterm_postinfix;
	}
op = XtrAtomIndex(tokenval);
if (is_postfix_op(op, &lprec, &oprec) &&
	(oprec <= prec) && (prec1 <= lprec))
	{
	PushCopyStackVar(*mptr, mp);
	if ((functor = find_func(op, (MW) 1)) == 0)
		PushFunctor(NullFunctor1);
	else
		PushFunctor(functor);
	*mptr = MakeStruct(copyptr);		     /* stackchange: no modification */
	prec1 = oprec;
	if (NotNextToken)
		{
		PushCopyStackVar(*mptr, mp);
		PushFunctor(ErrorFunctor1);
		*mptr = MakeStruct(copyptr);	      /* stackchange: no modification */
		return(ErrorNumber);
		}
	goto scan_subterm_postinfix;
	}
if (is_infix_op(op, &lprec, &oprec, &rprec) &&
	(oprec <= prec) && (prec1 <= lprec))
	{
	if (NotNextToken)
		{
		PushCopyStack(ErrorAtom);
		PushCopyStackVar(*mptr, mp);
		if ((functor = find_func(op, (MW) 2)) == 0)
			PushFunctor(NullFunctor2);
		else
			PushFunctor(functor);
		*mptr = MakeStruct(copyptr);	     /* stackchange: no modification */
		return(ErrorNumber);
		}
	ErrorNumber = scan_subterm(&m, rprec, follow);
	PushCopyStackVar(m, mp);
	PushCopyStackVar(*mptr, mp);
	if ((functor = find_func(op, (MW) 2)) == 0)
		PushFunctor(NullFunctor2);
	else
		PushFunctor(functor);
	*mptr = MakeStruct(copyptr);		    /* stackchange: no modification */
	if (ErrorNumber != Success)
		return(ErrorNumber);
	prec1 = oprec;
	goto scan_subterm_postinfix;
	}
return(Success);
}


/* search for errors in term mptr */
int scan_term(MWPTR mptr)
{
int     ErrorNumber;
FILE    *outfilesave;
MWPTR    mp;

ErrorFunctor1 = find_func(ErrorAtomIndex, (MW) 1);
ErrorFunctor2 = find_func(ErrorAtomIndex, (MW) 2);
NullFunctor1 = find_func(NullAtomIndex, (MW) 1);
NullFunctor2 = find_func(NullAtomIndex, (MW) 2);
CurlyFunctor2 = find_func(FunctIndex(CurlyBrackFunctor), (MW) 2);

varbase = NULL;
varlist = (MWPTR) trail;
readmode = TRUE;
if (NotCharPreRead)
	{
	SetCharPreRead;
	NextChar();
	}
if (NotNextToken)
	{
	m_error(ErrorNumber);
	return(ErrorNumber);
	}
if (scan_subterm(mptr, MaxPriority, FStopSet) != Success)
	{
scan_term_error:
	while ((token != FullStopToken) && (token != EofToken) &&
             ((ErrorNumber == SyntaxError) || ErrorNumber == OperatorError))
		(void) next_token();
	m_error(ErrorNumber);
	outfilesave = outfile;
	outfile = stderr;
	ErrorNumber = setjmp(io_err);
	if (ErrorNumber)
		{
		outfile = outfilesave;
		return(ErrorNumber);
		}
	write_term(mptr, MaxPriority, FALSE);
	}
if (token != FullStopToken)
	{
	ErrorNumber = SyntaxError;
	PushCopyStackVar(*mptr, mp);
	PushFunctor(ErrorFunctor1);
	*mptr = MakeStruct(copyptr);		 /* stackchange: no modification */
	goto scan_term_error;
	}
return(Success);
}

/* prints the term mptr into the current output file */
int display_term(MWPTR mptr)
{
MW      m;
MW      i;
MWPTR   functor;

m = *mptr;
StructDeref(m, mptr);
if (Nil(m))
	fprintf(outfile, "[]");
else if (Atom(m))
	fprintf(outfile, "%s", XtrAtomPtr(m));
else if (Int(m))
	fprintf(outfile, "%ld", (long) XtrInt(m));
else if (RefUndef(m))
	fprintf(outfile, "_%lx", (long) mptr);
else if (List(m))
	{
	fprintf(outfile, "[");
	for (;;)
		{
		mptr = XtrList(m);
		(void) display_term(mptr++);
		while(Ref(m = *mptr))
			mptr = XtrRef(m);
		if (List(m))
			fprintf(outfile, ",");
		else
			{
			if (NotNil(m))
				{
				fprintf(outfile, "|");
				(void) display_term(mptr);
				}
			fprintf(outfile, "]");
			return(Success);
			}
		}
	}
else
	{
	mptr = XtrStruct(m);
	functor = XtrFunctor(mptr);
	fprintf(outfile, "%s", IndexPtr(FunctIndex(functor)));
	i = FunctArity(functor);
	mptr = FirstArgPtr(mptr);
	fprintf(outfile, "(");
	while (--i > 0)
		{
		(void) display_term(mptr++);
		fprintf(outfile, ",");
		}
	(void) display_term(mptr);
	fprintf(outfile, ")");
	}
return(Success);
}


/* writes the atom at adress atomptr to the file 'outfile'
   if the parameter quoted is true --> the atom is put in quotes */
static void write_atom(CHARPTR atomptr, int quoted)
{
if (quoted)
	fprintf(outfile, "'%s'", atomptr);
else
	fprintf(outfile, "%s", atomptr);
}



/* writes the term mptr to the current putput file */
int write_term1(MWPTR mptr, int priority, int quoted, int level)
{
MW      m, op, i;
MWPTR   functor;
int     o_prior, l_prior, r_prior;
char    *cptr;

if (level < 0)
	{
	fprintf(outfile, "...");
	return(Success);
	}
m = *mptr;
StructDeref(m, mptr);
if (Nil(m))
	fprintf(outfile, "[]");
else if (Atom(m))
	if (m == ErrorAtom)
		{
		fprintf(outfile, "<<<< Error\n");
		longjmp(io_err, TRUE);
		}
	else
		write_atom(XtrAtomPtr(m), quoted);
else if (Int(m))
	fprintf(outfile, "%ld", (long) XtrInt(m));
else if (RefUndef(m))
	fprintf(outfile, "_%lx", (long) mptr);
else if (List(m))
	{
	fprintf(outfile, "[");
	for (;;)
		{
		mptr = XtrList(m);
		(void) write_term1(mptr++, MaxPriority, quoted, level-1);
		while(Ref(m = *mptr))
			mptr = XtrRef(m);
		if (List(m))
			fprintf(outfile, ",");
		else
			{
			if (NotNil(m))
				{
				fprintf(outfile, "|");
				(void) write_term1(mptr, MaxPriority, quoted, level);
				}
			fprintf(outfile, "]");
			return(Success);
			}
		}
	}
else
	{
	mptr = XtrStruct(m);
	functor = XtrFunctor(mptr);
	if (functor == CurlyBrackFunctor)
		{
		fprintf(outfile, "{");
		write_term1(FirstArgPtr(mptr), MaxPriority, quoted, level-1);
		fprintf(outfile, "}");
		return(Success);
		}	 
	if (functor == CurlyFunctor2)
		{
		fprintf(outfile, "{");
		write_term1(FirstArgPtr(mptr), MaxPriority, quoted, level-1);
		fprintf(outfile, ",");
		write_term1(FirstArgPtr(mptr) + 1, MaxPriority, quoted, level-1);
		fprintf(outfile, "}");
		return(Success);
		}	 
	op = FunctIndex(functor);
	i = FunctArity(functor);
	if ((i == 1) && (op == ErrorAtomIndex))
		{
		write_term1(FirstArgPtr(mptr), priority, quoted, level-1);
		fprintf(outfile, "<<<< Error\n");
		longjmp(io_err, TRUE);
		}
	cptr = IndexPtr(op);
	mptr = FirstArgPtr(mptr);
	if (i == 1)
		{
		if (is_prefix_op(op, &o_prior, &r_prior))
			{
			write_atom(cptr, quoted);
			fprintf(outfile, " ");
			if (o_prior > priority)
				{
				fprintf(outfile, "(");
				write_term1(mptr, MaxPriority, quoted, level-1);
				fprintf(outfile, ")");
				}
			else
				write_term1(mptr, r_prior, quoted, level-1);
			return(Success);
			}
		else if (is_postfix_op(op, &l_prior, &o_prior))
			{
			if (o_prior > priority)
				{
				fprintf(outfile, "(");
				write_term1(mptr, MaxPriority, quoted, level-1);
				fprintf(outfile, ")");
				}
			else
				write_term1(mptr, l_prior, quoted, level-1);
			fprintf(outfile, " ");
			write_atom(cptr, quoted);
			return(Success);
			}
		}
	else if ((i == 2) && is_infix_op(op, &l_prior, &o_prior, &r_prior))
		{
		if (o_prior > priority)
			{
			fprintf(outfile, "(");
			write_term1(mptr, MaxPriority, quoted, level-1);
			fprintf(outfile, " ");
			write_atom(cptr, quoted);
			fprintf(outfile, " ");
			write_term1(mptr+1, MaxPriority, quoted, level-1);
			fprintf(outfile, ")");
			}
		else
			{
			write_term1(mptr, l_prior, quoted, level-1);
			if (op == CommaAtomIndex)
				fprintf(outfile, ", ");
			else
				{
				fprintf(outfile, " ");
				write_atom(cptr, quoted);
				fprintf(outfile, " ");
				}
			write_term1(mptr+1, r_prior, quoted, level-1);
			}
		return(Success);
		}
	write_atom(cptr, quoted);
	fprintf(outfile, "(");
	while (--i > 0)
		{
		(void) write_term1(mptr++, MaxPriority, quoted, level-1);
		fprintf(outfile, ",");
		}
	(void) write_term1(mptr, MaxPriority, quoted, level-1);
	fprintf(outfile, ")");
	}
return(Success);
}


/* writes the term mptr to the current output file */
int write_term(MWPTR mptr, int priority, int quoted)
{
	return(write_term1( mptr, priority, quoted, WriteDepth));
}


/* opens arg as the current input file */
int b_see(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Atom(m))
	return(change_file(m, TRUE));
return(Fail);
}


/* opens arg as the current output file */
int b_tell(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if (Atom(m))
	return(change_file(m, FALSE));
return(Fail);
}


/* unifies the current input file with arg */
int b_seeing(MW arg[])
{
MW      m;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	*mptr = infile_name;
	return(Success);
	}
else if (m == infile_name)
	return(Success);
return(Fail);
}


/* unifies the file arg with the current output file */
int b_telling(MW arg[])
{
MW      m;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	*mptr = outfile_name;
	return(Success);
	}
else if (m == outfile_name)
	return(Success);
return(Fail);
}


/* closes the current input file */
int b_seen()
{
return(close_file(TRUE));
}


/* closes the current output file */
int b_told()
{
return(close_file(FALSE));
}


/* unifies the last read character with arg */
int b_get0(MW arg[])
{
MW      m;
MWPTR   mptr;

readmode = FALSE;
if (NotCharPreRead)
	NextChar();
if (cc == EM)
	return(Fail);
SetNoPreRead;
m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	*mptr = MakeInt(cread);
	return(Success);
	}
else if (m == (MW) MakeInt(cread))
	return(Success);
return(Fail);
}


/* reads the next character and unifies it with arg */
int b_get(MW arg[])
{
MW      m;
MWPTR   mptr;

readmode = FALSE;
if (NotCharPreRead)
	NextChar();
if (cc == EM)
	return(Fail);
while ((cc != EM) && (cread < ' ') && (cread > '~'))
	NextChar();
SetNoPreRead;
if (cc == EM)
	return(Fail);
m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	*mptr = MakeInt(cread);
	return(Success);
	}
else if (m == (MW) MakeInt(cread))
	return(Success);
return(Fail);

}


/* puts arg into the file outfile */
int b_put(MW arg[])
{
MW      m;

m = arg[0];
ValDeref(m);
if ((Int(m)) && (putc((int) XtrInt(m), outfile) != EOF))
	return(Success);
return(Fail);
}


/* reads the next term and unifies it with arg */
int b_read(MW arg[])
{
MW      m;
MWPTR   mptr;

m = *(mptr = arg);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	Trail(mptr);
	if (scan_term(mptr) == Success)
		return(Success);
	else
		return(Fail);
	}
if (scan_term(&m) != Success)
	return(Fail);
return(unify(arg, &m));
}


/* writes term arg[] (without quotes) */
int b_write(MW arg[])
{
return(write_term(arg, MaxPriority, FALSE));
}


/* writes term arg[] put in quotes */
int b_writeq(MW arg[])
{
return(write_term(arg, MaxPriority, TRUE));
}


/* display the term listed in arg[] */
int b_display(MW arg[])
{
return(display_term(arg));
}


/* builtin-predicate: reads until end of file */
int b_eof()
{
if (ComPreRead)
	return(Fail);
if (NotCharPreRead)
	{
	SetCharPreRead;
	NextChar();
	}
if (! readmode)
	if (cc == EM)
		return(Success);
	else
		return(Fail);

loop:
switch((int) cc)
	{
	case BS:
	case NL:
		NextChar();
		goto loop;
	case CM:
		for (;;)
			{
			NextChar();
			if (cc == EM)
				return(Success);
			if (cc == NL)
				{
				NextChar();
				goto loop;
				}
			}
	case C1:
		NextChar();
		if (cc != C2)
			{
			SetComPreRead;
			return(Fail);
			}
		for (;;)
			{
			NextChar();
			while (cc == C2)
				{
				NextChar();
				if (cc == C1)
					{
					NextChar();
					goto loop;
					}	       
				}
			if (cc == EM)
				return(ErrorNumber = EndOfInputError);
			}
	case EM:
		return(Success);
	default:
		return(Fail);
	}
}


/* builtin-predicate for printing a new line */
int b_nl()
{
#ifdef Atari
fprintf(outfile, "\015\n");
#else
fprintf(outfile, "\n");
#endif
return(Success);
}

